import sys
# # temp solution for directory.
sys.path.append("./src/")
from undefined.Utils import check_division_by_zero, check_pow
import math
import numpy as np



class UDFunction:
    def __init__(self, val, der=1):
        """
        This class is where we overload all the operators, which will be used to calculate the derivatives.

        Args:
            val (numeric or numpy ndarray): value of function
            der (int, optional): derivative of function. Defaults to 1.
        """
        self._val = val
        self._der = der
        if hasattr(self._val, 'shape'):
            self._shape = self._val.shape
        else:
            self._shape = 1
        self._left_child = None
        self._right_child = None

    @property
    def val(self):
        """
        This is a decorator return rouded input self.val

        Returns:
            array: 2 decimal rounded input of self.value
        """
        if isinstance(self._val, float):
            return round(self._val, 2)
        elif isinstance(self._val, np.ndarray):
            return np.round(self._val, 2)
        else:
            return self._val

    @property
    def der(self):
        """
        This is a decorator return rouded input self.der

        Returns:
            array: 3 decimal rounded input of self.der
        """
        if isinstance(self._der, float):
            return round(self._der, 3)
        elif isinstance(self._der, np.ndarray):
            return np.round(self._der, 3)
        else:
            return self._der

    def __str__(self):
        return f"value: {self.val} \n" + f"derivative: {self.der}"

    def __add__(self, other):
        """
        This allows to do addition with UDFunction instances or scalar numbers, and calculate the value after taking the derivative.
        TypeError will raise if none of the self or other are UDFunction instances.

        Args:
            other (UDFunction or numeric): object to add with
        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            new_val = self._val + other._val
            new_der = self._der + other._der
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val + other
            new_der = self._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __mul__(self, other):
        """
        This allows to do multification with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative.
        TypeError will raise if none of the self or other are UDFunction instances.

        Args:
            other (UDFunction or numeric): object to multiply with
        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            new_val = self._val * other._val
            new_der = self._der * other._val + self._val * other._der
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val * other
            new_der = self._der * other
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __radd__(self, other):
        """
        This is called when int/float or UDFunction instances + an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to add with
        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            new_val = self._val + other._val
            new_der = self._der + other._der
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val + other
            new_der = self._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __rmul__(self, other):
        """
        This is called when int/float or UDFunction instances * an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to multiply with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            new_val = self._val * other._val
            new_der = self._der * other._val + self._val * other._der
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val * other
            new_der = self._der * other
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __neg__(self):
        """
        This allows to negate UDFunction instances itself.

        Returns:
            UDFunction: object with neg value
        """
        return -1 * self

    def __sub__(self, other):
        """
        This allows to do subtraction with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative.
        TypeError will raise if none of the self or other are UDFunction instances.

        Args:
            other (UDFunction or numeric): object to subtract with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            new_val = self._val - other._val
            new_der = self._der - other._der
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val - other
            new_der = self._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __rsub__(self, other):
        """
        This is called when int/float or UDFunction instances - an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to subtract with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            new_val = other._val - self._val
            new_der = other._der - self._der
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = other - self._val
            new_der = - self._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __truediv__(self, other):
        """
        This allows to do true division with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative.
        TypeError will raise if none of the self or other are UDFunction instances.

        Args:
            other (UDFunction or numeric): object to (true) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            check_division_by_zero(other._val)
            new_val = self._val / other._val
            new_der = (self._der * other._val - self._val *
                       other._der) / (other._val * other._val)
        elif isinstance(other, (int, float, np.ndarray)):
            check_division_by_zero(other)
            new_val = self._val / other
            new_der = self._der / other
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __rtruediv__(self, other):
        """
        This is called when int/float or UDFunction instances / (divide) an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to (true) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        check_division_by_zero(self._val)
        if isinstance(other, UDFunction):
            new_val = other._val / self._val
            new_der = (self._val * other._der - self._der *
                       other._val) / (self._val * self._val)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = other / self._val
            new_der = - 1 * other * self._der / (self._val * self._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __floordiv__(self, other):  # self // other
        """
        This allows to do floor division with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative.
        TypeError will raise if none of the self or other are UDFunction instances.

        Args:
            other (UDFunction or numeric): object to (floor) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDFunction):
            check_division_by_zero(other._val)
            new_val = self._val // other._val
            new_der = (self._der * other._val - self._val *
                       other._der) // (other._val * other._val)
        elif isinstance(other, (int, float, np.ndarray)):
            check_division_by_zero(other)
            new_val = self._val // other
            new_der = self._der // other
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __rfloordiv__(self, other):
        """
        This is called when int/float or UDFunction instances // (floor divide) an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to (floor) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        check_division_by_zero(self._val)
        if isinstance(other, UDFunction):
            new_val = other._val // self._val
            new_der = (self._val * other._der - self._der *
                       other._val) // (self._val * self._val)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = other // self._val
            new_der = - 1 * other * self._der // (self._val * self._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    def __pow__(self, other):
        """
        This allows to do "to the power" with UDFunction instances or scalar numbers, and calculate the value after taking the derivative.
        ** operator.

        Args:
            other (any): object to take power of.

        Returns:
            UDFunction: a new object with new_val and new_der
        """

        if isinstance(other, UDFunction):
            check_pow(self._val, other._val)
            if isinstance(self._val, (int, float)):
                new_val = self._val ** other._val
                if isinstance(other._val, np.ndarray):
                    new_der_1 = other._val * \
                        np.power(self._val, other._val - 1) * self._der
                    new_der_2 = math.log(self._val) * new_val * other._der
                    new_der = new_der_1 + new_der_2
                else:
                    new_der_1 = other._val * \
                        self._val ** (other._val - 1) * self._der
                    new_der_2 = math.log(self._val) * new_val * other._der
                    new_der = new_der_1 + new_der_2
            else:  # self._val is of type ndarray
                if isinstance(other._val, np.ndarray):
                    if other._val.shape[0] != self._val.shape[0]:
                        raise ValueError(
                            f"error raised by undefined: operands could not be broadcast together with shapes {other._val.shape} {self._val.shape}")
                    else:
                        new_val = self._val ** other._val
                        new_der_1 = other._val * \
                            np.power(self._val, other._val - 1) * self._der
                        new_der_2 = np.log(self._val) * \
                            new_val * other._der
                        new_der = new_der_1 + new_der_2
                else:
                    new_val = self._val ** other._val
                    new_der_1 = other._val * \
                        self._val ** (other._val - 1) * self._der
                    new_der_2 = np.log(self._val) * new_val * other._der
                    new_der = new_der_1 + new_der_2
        elif isinstance(other, (int, float, np.ndarray)):
            check_pow(self._val, other)
            if isinstance(self._val, np.ndarray):
                new_val = np.power(self._val, other)
                new_der = other * \
                    np.power(self._val, other - 1) * self._der
            elif isinstance(self._val, (int, float)):
                new_val = self._val ** other
                new_der = other * self._val**(other - 1) * self._der
        return UDFunction(new_val, new_der)

    def __rpow__(self, other):
        """
        This allows to do "to the power" with UDFunction instances or scalar numbers, and calculate the value after taking the derivative.
        ** operator.
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            degree (numeric): object to take power of.

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        # other ^ self
        if isinstance(other, UDFunction):
            check_pow(other._val, self._val)
            if isinstance(other._val, (int, float)):
                new_val = other._val ** self._val
                new_der_1 = np.log(other._val) * new_val * self._der
                new_der_2 = self._val * \
                    other._val ** (self._val - 1) * other._der
                new_der = new_der_1 + new_der_2
            else:
                if isinstance(self._val, np.ndarray):
                    if other._val.shape[0] != self._val.shape[0]:
                        raise ValueError(
                            f"error raised by undefined: operands could not be broadcast together with shapes {other._val.shape} {self._val.shape}")
                    else:
                        new_val = other._val ** self._val
                        new_der_1 = np.log(other._val) * new_val * self._der
                        new_der_2 = self._val * \
                            np.power(other._val, (self._val - 1)) * other._der
                else:
                    new_val = other._val ** self._val
                    new_der_1 = np.log(other._val) * new_val * self._der
                    new_der_2 = self._val * \
                        other._val ** (self._val - 1) * other._der

        elif isinstance(other, (int, float, np.ndarray)):
            check_pow(other, self._val)
            new_val = other ** self._val
            new_der = math.log(other) * new_val * self._der

        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")

        return UDFunction(new_val, new_der)

    def __eq__(self, other):
        """compare whether the two UDFunction objects have the same values.
        Return true if equal, and false otherwise.

        raise TypeError is other is not a UDFunction object.

        Args:
            other ([UDFunction])

        Returns:
            True if equal. Otherwise False
        """
        if isinstance(other, UDFunction):
            return self.val == other.val
        elif isinstance(other, (int, float)):
            return self.val == other
        else:
            raise TypeError("error raised by undefined: Need a UDFunction object to compare")

    def __ne__(self, other):
        """compare whether the two UDFunction objects have different values.
        raise TypeError is other is not a UDFunction object.

        Args:
            other ([UDFunction])

        Returns:
            True if not equal. Otherwise False
        """
        if isinstance(other, UDFunction):
            return self.val != other.val
        elif isinstance(other, (int, float)):
            return self.val != other
        else:
            raise TypeError("error raised by undefined: Need a UDFunction object to compare")

    def __lt__(self, other):
        """overload the < operator
        raise TypeError is other is not a UDFunction object.

        Args:
            other ([UDFunction])

        Returns:
            True if less than. Otherwise False
        """
        if isinstance(other, UDFunction):
            return self.val < other.val
        elif isinstance(other, (int, float)):
            return self.val < other
        else:
            raise TypeError("error raised by undefined: Need a UDFunction object to compare")

    def __gt__(self, other):
        """overload the > operator
        raise TypeError is other is not a UDFunction object.

        Args:
            other ([UDFunction])
        
        Returns:
            True if greater than. Otherwise False
        """
        if isinstance(other, UDFunction):
            return self.val > other.val
        elif isinstance(other, (int, float)):
            return self.val > other
        else:
            raise TypeError("error raised by undefined: Need a UDFunction object to compare")

    def __le__(self, other):
        """overload the > operator
        raise TypeError is other is not a UDFunction object.

        Args:
            other ([UDFunction])

        Returns:
            True if less than or equal to. Otherwise False
        """
        if isinstance(other, UDFunction):
            return self.val <= other.val
        elif isinstance(other, (int, float)):
            return self.val <= other
        else:
            raise TypeError("error raised by undefined: Need a UDFunction object to compare")

    def __ge__(self, other):
        """overload the > operator
        raise TypeError is other is not a UDFunction object.

        Args:
            other ([UDFunction])
        
        Returns:
            True if greater than or equal to. Otherwise False
        """
        if isinstance(other, UDFunction):
            return self.val >= other.val
        elif isinstance(other, (int, float)):
            return self.val >= other
        else:
            raise TypeError("error raised by undefined: Need a UDFunction object to compare")
    
    def __round__(self, digit):
        '''overwrite the round method.
            Internal method for testing.
        '''
        return round(self.val, digit)




import sys
sys.path.append("./src/")
import unittest
from datetime import datetime
from undefined.Utils import time

class TestUDFunction(unittest.TestCase):
    def test_time(self):
        now = datetime.now()
        self.assertEqual(time()[:8], now.strftime("%H:%M:%S:%f")[:8])


import sys
sys.path.append("./src/")
import numpy as np
import math
from undefined.GraphGenerator import UDGraph
from undefined.UDFunction import UDFunction
from undefined import Calculator as cal
from typing import Type
import unittest



class TestCalculator(unittest.TestCase):
    def setUp(self):
        self.a = 2
        self.b = - 1.11
        self.c = 99999.99
        self.d = -99999.99
        self.e = np.array([[self.a, self.b]])
        self.f = np.array([[self.a, self.b, self.c, self.d]])
        self.g = math.pi / 2
        self.h = math.e
        self.j = np.array([[self.a, self.b, self.g, self.h]])
        self.s = "str"

        self.af = UDFunction(self.a)
        self.bf = UDFunction(self.b)
        self.cf = UDFunction(self.c)
        self.df = UDFunction(self.d)
        self.ef = UDFunction(self.e)
        self.ff = UDFunction(self.f)
        self.gf = UDFunction(self.g)
        self.hf = UDFunction(self.h)
        self.jf = UDFunction(self.j)

        self.ag = UDGraph(self.a)
        self.bg = UDGraph(self.b)
        self.cg = UDGraph(self.c)
        self.dg = UDGraph(self.d)
        self.eg = UDGraph(self.e)
        self.fg = UDGraph(self.f)
        self.gg = UDGraph(self.g)
        self.hg = UDGraph(self.h)
        self.jg = UDGraph(self.j)

        x = UDFunction(1.0)
        # test cosine
        self.f1 = cal.cos(x)
        self.f2 = cal.cos(2*x + 1)

        # test sine
        self.f3 = cal.sin(x)
        self.f4 = cal.sin(2*x + 2)

        # test tangent
        self.f5 = cal.tan(x)
        self.f6 = cal.tan(2*x + 2)

        # test sqrt
        self.f7 = cal.sqrt(x)
        self.f8 = 2*x + cal.sqrt(2*x)

        # test exp
        self.f9 = cal.exp(x)
        self.f10 = 2*x - cal.exp(x) + 1

        # test log
        self.f11 = cal.log(x, math.e)
        self.f12 = 5*x + cal.log(x+1, math.e) - 2/x

        self.f13 = cal.log(x, 2)
        self.f1000 = cal.log(x, np.e)

    def assertListAlmostEqual(self, o1, o2):
        for i, o in enumerate(o1):
            if round(o, 2) != o2[i]:
                return False
        return True

    def test_cos(self):
        self.assertEqual(round(cal.cos(self.a), 2), - 0.42)
        self.assertEqual(round(cal.cos(self.b), 2), 0.44)
        self.assertEqual(round(cal.cos(self.c), 2), - 1.00)
        self.assertEqual(round(cal.cos(self.d), 2), - 1.00)

        self.assertListAlmostEqual(cal.cos(self.e)[0], [-0.42, 0.44])
        self.assertListAlmostEqual(cal.cos(self.f)[0], [-1.00, -1.00])

        self.assertEqual(round(cal.cos(self.g), 2), 0.00)
        self.assertEqual(round(cal.cos(self.h), 2), -0.91)
        self.assertListAlmostEqual(
            cal.cos(self.j)[0], [-0.42, 0.44, 0.00, -0.91])

        with self.assertRaises(TypeError):
            cal.cos(self.s)

        self.assertEqual(cal.cos(self.af).val, - 0.42)
        self.assertEqual(cal.cos(self.bf).val, 0.44)
        self.assertEqual(cal.cos(self.cf).val, - 1.00)
        self.assertEqual(cal.cos(self.df).val, - 1.00)

        self.assertListAlmostEqual(cal.cos(self.ef).val[0], [-0.42, 0.44])
        self.assertListAlmostEqual(cal.cos(self.ff).val[0], [-1.00, -1.00])

        self.assertEqual(cal.cos(self.gf).val, 0.00)
        self.assertEqual(cal.cos(self.hf).val, -0.91)
        self.assertListAlmostEqual(
            cal.cos(self.jf).val[0], [-0.42, 0.44, 0.00, -0.91])

        self.assertEqual(cal.cos(self.ag).val, - 0.42)
        self.assertEqual(cal.cos(self.bg).val, 0.44)
        self.assertEqual(cal.cos(self.cg).val, - 1.00)
        self.assertEqual(cal.cos(self.dg).val, - 1.00)

        self.assertListAlmostEqual(cal.cos(self.eg).val[0], [-0.42, 0.44])
        self.assertListAlmostEqual(cal.cos(self.fg).val[0], [-1.00, -1.00])

        self.assertEqual(cal.cos(self.gg).val, 0.00)
        self.assertEqual(cal.cos(self.hg).val, -0.91)
        self.assertListAlmostEqual(
            cal.cos(self.jg).val[0], [-0.42, 0.44, 0.00, -0.91])

    def test_sin(self):
        self.assertEqual(round(cal.sin(self.a), 2), 0.91)
        self.assertEqual(round(cal.sin(self.b), 2), - 0.90)
        self.assertEqual(round(cal.sin(self.c), 2), 0.05)
        self.assertEqual(round(cal.sin(self.d), 2), - 0.05)

        self.assertListAlmostEqual(cal.sin(self.e)[0], [0.91, - 0.90])
        self.assertListAlmostEqual(cal.sin(self.f)[0], [0.05, -0.05])
        self.assertEqual(round(cal.sin(self.g), 2), 1.00)
        self.assertEqual(round(cal.sin(self.h), 2), 0.41)
        self.assertListAlmostEqual(
            cal.sin(self.j)[0], [-0.42, 0.44, 1.00, 0.41])

        with self.assertRaises(TypeError):
            cal.sin(self.s)

        self.assertEqual(cal.sin(self.af).val, 0.91)
        self.assertEqual(cal.sin(self.bf).val, - 0.90)
        self.assertEqual(cal.sin(self.cf).val, 0.05)
        self.assertEqual(cal.sin(self.df).val, - 0.05)

        self.assertListAlmostEqual(cal.sin(self.ef).val[0], [0.91, - 0.90])
        self.assertListAlmostEqual(cal.sin(self.ff).val[0], [0.05, -0.05])
        self.assertEqual(cal.sin(self.gf).val, 1.00)
        self.assertEqual(cal.sin(self.hf).val, 0.41)
        self.assertListAlmostEqual(
            cal.sin(self.jf).val[0], [-0.42, 0.44, 1.00, 0.41])

        self.assertEqual(cal.sin(self.ag).val, 0.91)
        self.assertEqual(cal.sin(self.bg).val, - 0.90)
        self.assertEqual(cal.sin(self.cg).val, 0.05)
        self.assertEqual(cal.sin(self.dg).val, - 0.05)

        self.assertListAlmostEqual(cal.sin(self.eg).val[0], [0.91, - 0.90])
        self.assertListAlmostEqual(cal.sin(self.fg).val[0], [0.05, -0.05])
        self.assertEqual(cal.sin(self.gg).val, 1.00)
        self.assertEqual(cal.sin(self.hg).val, 0.41)
        self.assertListAlmostEqual(
            cal.sin(self.jg).val[0], [-0.42, 0.44, 1.00, 0.41])

    def test_sqrt(self):
        self.assertEqual(round(cal.sqrt(self.a), 2), 1.41)
        with self.assertRaises(ValueError):
            cal.sqrt(self.b)
        with self.assertRaises(ValueError):
            cal.sqrt(self.e)
        with self.assertRaises(ValueError):
            cal.sqrt(self.f)

        with self.assertRaises(TypeError):
            cal.sqrt(self.s)
        self.assertEqual(round(cal.sqrt(self.g), 2), 1.25)

        self.assertEqual(cal.sqrt(self.af).val,  1.41)
        with self.assertRaises(ValueError):
            cal.sqrt(self.bf)

        with self.assertRaises(ValueError):
            cal.sqrt(self.ef)
        with self.assertRaises(ValueError):
            cal.sqrt(self.ff)
        self.assertEqual(cal.sqrt(self.ag).val,  1.41)
        with self.assertRaises(ValueError):
            cal.sqrt(self.bg)
        with self.assertRaises(ValueError):
            cal.sqrt(self.eg)
        with self.assertRaises(ValueError):
            cal.sqrt(self.fg)

    def test_tan(self):
        self.assertEqual(round(cal.tan(self.a), 2), -2.19)
        self.assertEqual(round(cal.tan(self.b), 2), - 2.01)
        self.assertEqual(round(cal.tan(self.c), 2), - 0.05)
        self.assertEqual(round(cal.tan(self.d), 2), 0.05)

        self.assertListAlmostEqual(cal.tan(self.e)[0], [-2.19, - 2.01])
        self.assertListAlmostEqual(cal.tan(self.f)[0], [- 0.05, 0.05])
        with self.assertRaises(TypeError):
            cal.tan(self.s)
        with self.assertRaises(ZeroDivisionError):
            cal.tan(self.g)
        self.assertEqual(round(cal.tan(self.h), 2), -0.45)
        with self.assertRaises(ZeroDivisionError):
            cal.tan(self.j)

        self.assertEqual(cal.tan(self.af).val,  -2.19)
        self.assertEqual(cal.tan(self.bf).val, - 2.01)
        self.assertEqual(cal.tan(self.cf).val, - 0.05)
        self.assertEqual(cal.tan(self.df).val, 0.05)

        self.assertListAlmostEqual(cal.tan(self.ef).val[0], [-2.19, - 2.01])
        self.assertListAlmostEqual(cal.tan(self.ff).val[0], [- 0.05, 0.05])
        with self.assertRaises(ZeroDivisionError):
            cal.tan(self.gf)
        self.assertEqual(cal.tan(self.hf).val, -0.45)
        with self.assertRaises(ZeroDivisionError):
            cal.tan(self.jf)

        self.assertEqual(cal.tan(self.ag).val,  -2.19)
        self.assertEqual(cal.tan(self.bg).val, - 2.01)
        self.assertEqual(cal.tan(self.cg).val, - 0.05)
        self.assertEqual(cal.tan(self.dg).val, 0.05)

        self.assertListAlmostEqual(cal.tan(self.eg).val[0], [-2.19, - 2.01])
        self.assertListAlmostEqual(cal.tan(self.fg).val[0], [- 0.05, 0.05])
        with self.assertRaises(ZeroDivisionError):
            cal.tan(self.gg)
        self.assertEqual(cal.tan(self.hg).val, -0.45)
        with self.assertRaises(ZeroDivisionError):
            cal.tan(self.jg)

    def test_log(self):
        self.assertEqual(round(cal.log(self.a), 2), 0.69)
        self.assertEqual(round(cal.log(self.a, 10), 2), 0.30)
        with self.assertRaises(ValueError):
            cal.log(self.b)
        with self.assertRaises(ValueError):
            cal.log(self.e)
        with self.assertRaises(ValueError):
            cal.log(self.f)

        with self.assertRaises(TypeError):
            cal.log(self.s)
        self.assertEqual(round(cal.log(self.g), 2), 0.45)
        self.assertEqual(cal.log(self.af).val,  0.69)
        self.assertEqual(cal.log(self.af, 10).val,  0.30)
        with self.assertRaises(ValueError):
            cal.log(self.bf)

        with self.assertRaises(ValueError):
            cal.log(self.ef)
        with self.assertRaises(ValueError):
            cal.log(self.ff)
        self.assertEqual(cal.log(self.ag).val,  0.69)
        self.assertEqual(cal.log(self.ag, 10).val,  0.30)
        with self.assertRaises(ValueError):
            cal.log(self.bg)
        with self.assertRaises(ValueError):
            cal.log(self.eg)
        with self.assertRaises(ValueError):
            cal.log(self.fg)
        with self.assertRaises(ValueError):
            cal.log(self.fg, -1)

    def test_sine_integration(self):

        a = "2.0"
        x = UDFunction(a)

        self.assertEqual(round(self.f3.val, 2), 0.84)
        self.assertEqual(round(self.f3.der, 2), 0.54)

        self.assertEqual(round(self.f4.val, 2), -0.76)
        self.assertEqual(round(self.f4.der, 2), -1.31)

        with self.assertRaises(TypeError):
            self.f45 = cal.sin(2*x) + x

        self.assertEqual(cal.sin(np.array([np.pi/2])), 1)
        self.assertEqual(cal.sin(np.pi/2), 1)

        with self.assertRaises(TypeError):
            self.f000 = cal.sin("3/np.pi")

        # UDgraph
        self.f252 = cal.sin(UDGraph(np.array([np.pi/2])))

        # need to check the calculation
        self.assertEqual(np.round(self.f252.val), np.array([1.]))

        with self.assertRaises(TypeError):
            cal.sin(UDGraph("3/np.pi"))

    def test_cosine_integration(self):
        a = "2.0"
        x = UDFunction(a)

        self.assertEqual(round(self.f1.val, 2), 0.54)
        self.assertEqual(round(self.f1.der, 2), -0.84)

        self.assertEqual(round(self.f2.val, 2), -0.99)
        self.assertEqual(round(self.f2.der, 2), -0.28)

        with self.assertRaises(TypeError):
            self.f46 = cal.cos(2*x) + x

        self.assertEqual(cal.cos(np.array([np.pi])), -1)
        self.assertEqual(cal.cos(np.pi), -1)

        with self.assertRaises(TypeError):
            self.f000 = cal.cos("3/np.pi")

        # UDgraph input

        self.f250 = cal.cos(UDGraph(2))
        self.f251 = cal.cos(UDGraph(np.array([2])))
        self.assertEqual(round(self.f250.val, 2), -0.42)
        self.assertEqual(self.f251.val, np.array([-0.42]))

        with self.assertRaises(TypeError):
            self.f0001 = cal.cos(UDGraph(a))

    def test_tangent_integration(self):

        a = "2.0"
        x = UDFunction(a)

        self.assertEqual(round(self.f5.val, 2), 1.56)
        self.assertEqual(round(self.f5.der, 2), 3.43)

        self.assertEqual(round(self.f6.val, 2), 1.16)
        self.assertEqual(round(self.f6.der, 2), 4.68)

        with self.assertRaises(TypeError):
            self.f47 = cal.tan(2*x) + x

        self.assertEqual(cal.tan(np.array([0])), 0)
        self.assertEqual(cal.tan(0), 0)

        with self.assertRaises(TypeError):
            self.f000 = cal.tan("3/np.pi")

        # UDgraph input

        self.f250 = cal.tan(UDGraph(2))
        self.f251 = cal.tan(UDGraph(np.array([2])))
        self.assertEqual(round(self.f250.val, 2), -2.19)
        self.assertEqual(self.f251.val, np.array([-2.19]))

        with self.assertRaises(TypeError):
            self.f0001 = cal.tan(UDGraph(a))

    def test_sqrt_integration(self):
        a = "2.0"
        x = UDFunction(a)
        y = UDFunction(UDGraph("abc"))

        self.assertEqual(round(self.f7.val, 2), 1)
        self.assertEqual(round(self.f7.der, 2), 0.5)

        self.assertEqual(round(self.f8.val, 2), 3.41)
        self.assertEqual(round(self.f8.der, 2), 2.71)

        with self.assertRaises(TypeError):
            self.f48 = cal.sqrt(2*x) + x

        with self.assertRaises(TypeError):
            cal.sqrt(2*x) + x

        with self.assertRaises(TypeError):
            cal.sqrt(y)

        self.assertEqual(cal.sqrt(np.array([4])), 2)
        self.assertEqual(cal.sqrt(4), 2)

        with self.assertRaises(TypeError):
            self.f000 = cal.sqrt("3/np.pi")

    def test_exp_integration(self):

        a = "2.0"
        x = UDFunction(a)
        x1 = UDFunction(1)
        y = UDGraph(np.array([1]))
        f_logist = cal.standard_logistic(x1)

        self.assertEqual(f_logist.val, 0.73)
        self.assertEqual(round(self.f9.val, 2), 2.72)
        self.assertEqual(round(self.f9.der, 2), 2.72)

        self.assertEqual(round(self.f10.val, 2), 0.28)
        self.assertEqual(round(self.f10.der, 2), -0.72)

        self.assertEqual(np.round(cal.exp(y).val, 2), 2.72)

        with self.assertRaises(TypeError):
            cal.exp(UDGraph("str"))

        with self.assertRaises(TypeError):
            self.f49 = cal.exp(2*x) + x

        self.assertEqual(cal.exp(np.array([1])), np.e)
        self.assertEqual(cal.exp(1), np.e)

        with self.assertRaises(TypeError):
            self.f000 = cal.exp("3/np.pi")

    def test_log_integration(self):
        a = "2.0"
        x = UDFunction(a)
        y = UDGraph(1)
        z = UDGraph(np.array([1]))
        b = UDGraph("a")

        self.assertEqual(round(self.f11.val, 2), 0)
        self.assertEqual(round(self.f11.der, 2), 1)

        self.assertEqual(round(self.f12.val, 2), 3.69)
        self.assertEqual(round(self.f12.der, 2), 7.5)

        self.assertEqual(round(self.f13.val, 2), 0)
        self.assertEqual(round(self.f13.der, 2), 1.44)

        with self.assertRaises(TypeError):
            self.f50 = cal.log(2*x, 2) + x

        self.assertEqual(cal.log(np.array([1]), np.e), 0)
        self.assertEqual(cal.log(1, np.e), 0)

        with self.assertRaises(TypeError):
            self.f000 = cal.log("3/np.pi", np.e)

        self.assertEqual(np.round(cal.log(y).val, 2), 0.0)
        self.assertEqual(np.round(cal.log(z).val, 2), np.array([0.]))

        with self.assertRaises(TypeError):
            self.f000 = cal.log(b)

    def test_hs_integration(self):
        a = 1
        x = UDFunction(a)

        self.f255 = cal.sinh(x)

        self.assertEqual(self.f255.val, 1.18)

        self.f255 = cal.cosh(x)

        self.assertEqual(self.f255.val, 1.54)

        self.f255 = cal.tanh(x)

        self.assertEqual(self.f255.val, 0.76)

        self.f255 = cal.coth(x)

        self.assertEqual(self.f255.val, 1.31)

        self.f255 = cal.sech(x)

        self.assertEqual(self.f255.val, 0.65)

        self.f255 = cal.csch(x)

        self.assertEqual(self.f255.val, 0.85)

        # test logistic

        self.f255 = cal.standard_logistic(x)

        self.assertEqual(round(self.f255.val, 2), 0.73)

    def test_arc(self):

        a = 0.5
        x = UDFunction(a)
        b = "0.5"
        y = UDFunction(b)
        # test arccos
        c = 99
        d = -99
        e = -1
        z1 = UDFunction(c)
        z2 = UDFunction(np.array([[a, c, d]]))
        z3 = UDGraph(c)
        z4 = UDGraph(np.array([[a, c, e]]))
        self.assertEqual(round(cal.arccos(x), 2), 1.05)
        self.assertEqual(round(cal.arccos(0.5), 2), 1.05)
        self.assertEqual(
            np.round(cal.arccos(UDFunction(np.array([0.5]))).val), 1)
        self.assertEqual(np.round(cal.arccos(np.array([0.5]))), 1)

        with self.assertRaises(TypeError):
            cal.arccos(y)
        with self.assertRaises(ValueError):
            cal.arccos(c)
        with self.assertRaises(ValueError):
            cal.arccos(z1)
        with self.assertRaises(ValueError):
            cal.arccos(z2)
        with self.assertRaises(ValueError):
            cal.arccos(z4)
        with self.assertRaises(ValueError):
            cal.arcsin(c)
        with self.assertRaises(ValueError):
            cal.arcsin(z1)
        with self.assertRaises(ValueError):
            cal.arcsin(z2)
        with self.assertRaises(ValueError):
            cal.arcsin(z3)
        with self.assertRaises(ValueError):
            cal.arcsin(z4)
        with self.assertRaises(TypeError):
            self.f000 = cal.arccos("3/np.pi")

        # UDgraph
        self.f256 = cal.arccos(UDGraph(0.5))
        self.f252 = cal.arccos(UDGraph(np.array([0.5])))

        # need to check the calculation
        self.assertEqual(round(self.f256.val), np.array([1.]))
        self.assertEqual(np.round(self.f252.val), np.array([1.]))

        with self.assertRaises(TypeError):
            cal.arccos(UDGraph("3/np.pi"))

        # test arcsin

        self.assertEqual(round(cal.arcsin(x), 2), 0.52)
        self.assertEqual(round(cal.arcsin(0.5), 2), 0.52)
        self.assertEqual(
            np.round(cal.arcsin(UDFunction(np.array([0.5]))).val), 1)
        self.assertEqual(np.round(cal.arcsin(np.array([0.5]))), 1)

        with self.assertRaises(TypeError):
            cal.arcsin(y)

        with self.assertRaises(TypeError):
            self.f000 = cal.arcsin("3/np.pi")

        # UDgraph
        self.f256 = cal.arcsin(UDGraph(0.5))
        self.f252 = cal.arcsin(UDGraph(np.array([0.5])))

        # need to check the calculation
        self.assertEqual(round(self.f256.val), np.array([1.]))
        self.assertEqual(np.round(self.f252.val), np.array([1.]))

        with self.assertRaises(TypeError):
            cal.arcsin(UDGraph("3/np.pi"))

        # test arctan

        self.assertEqual(round(cal.arctan(x), 2), 0.46)
        self.assertEqual(round(cal.arctan(0.5), 2), 0.46)
        self.assertEqual(np.round(cal.arctan(UDFunction(
            np.array([0.5]))).val, 2), np.array([0.46]))
        self.assertEqual(
            np.round(cal.arctan(np.array([0.5])), 2), np.array([0.46]))

        with self.assertRaises(TypeError):
            cal.arctan(y)

        with self.assertRaises(TypeError):
            self.f000 = cal.arctan("3/np.pi")

        # UDgraph
        self.f256 = cal.arctan(UDGraph(0.5))
        self.f252 = cal.arctan(UDGraph(np.array([0.5])))

        # need to check the calculation
        self.assertEqual(round(self.f256.val, 2), 0.46)
        self.assertEqual(np.round(self.f252.val, 2), 0.46)

        with self.assertRaises(TypeError):
            cal.arctan(UDGraph("3/np.pi"))


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from datetime import datetime
import numpy as np


class UDPrimitive(Enum):
    VAR = 0
    ADD = 1
    RADD = 2
    MUL = 3
    RMUL = 4
    NEG = 5
    SUB = 6
    RSUB = 7
    TRUEDIV = 8
    RTRUEDIV = 9
    FLOORDIV = 10
    RFLOORDIV = 11
    POW = 12
    RPOW = 13
    COS = 14
    SIN = 15
    TAN = 16
    SQRT = 17
    EXP = 18
    LOG = 19
    ACOS = 20
    ASIN = 21
    ATAN = 22


def time():
    """get current time at the most precision in seconds

    Returns:
        return the current time in string in Hour:Minute:Seconds:precise time format
    """
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S:%f")
    return current_time


def check_division_by_zero(val):
    """Used to check if unlawful division is executed by users
    raise errors if happened.

    Args:
        val ([int or float])

    Raises:
        ZeroDivisionError raised if dividing by 0.
    """
    if isinstance(val, np.ndarray):
        if not np.all(np.round(val, 7)):
            raise ZeroDivisionError("error raised by undefined: divide by zero encountered")
    elif isinstance(val, (int, float)):
        if round(val,7) == 0:
            raise ZeroDivisionError("error raised by undefined: divide by zero encountered")


def check_pow(val, degree):
    """Used to check if unlawful power is executed by users
    raise errors if happened.

    Args:
        val ([int or float])
        degree ([int or float])

    Raises:
        raise ValueError if any of the val or degree is NA type.
    """
    try:
        temp = np.power(val, degree)
        if np.any(np.isnan(temp)):
            raise ValueError("error raised by undefined: invalid inputs for pow()")
    except ValueError as e:
        raise ValueError(f"error raised by undefined: {e}")


def check_log(val, base):
    """Used to check if unlawful log is executed by users
    raise errors if happened. Cannot take log for number that is less or equal to 0
    log base also needs to be greater than 0.

    Args:
        val ([int or float])
        base ([int or float])

    Raises:
        raise ValueError if any of the base or val is less than 0.
    """
    if base <= 0:
        raise ValueError(f"error raised by undefined: invalid base {base} for log")
    if isinstance(val, np.ndarray):
        if np.any(val <= 0):
            raise ValueError(f"error raised by undefined: invalid value {val} for log")
    elif isinstance(val, (int, float)):
        if val <= 0:
            raise ValueError(f"error raised by undefined: invalid value {val} for log")


def check_arc(val):
    """Used to check if unlawful inverse trig function is executed by users
    raise errors if happened. Cannot take inverse trig function that is not between -1 and 1

    Args:
        val ([int or float])

    Raises:
        raise error if number is not between -1 and 1
    """
    if isinstance(val, np.ndarray):
        if not np.all(val > -1 and val < 1):
            raise ValueError(
                f"error raised by undefined: invalid values {val}, which should all be within (-1, 1)")
    elif isinstance(val, (int, float)):
        if not (val > -1 and val < 1):
            raise ValueError(
                f"error raised by undefined: invalid value {val}, which should be within (-1, 1)")




# we are going to test the forward model, reverse model and multiple vector functions here.
import sys
sys.path.append("./src/")
import numpy as np
from undefined.GraphGenerator import UDGraph, GeneratorHelper
from undefined.Calculator import sin, cos, exp, tan, sqrt, log
from undefined.UDFunction import UDFunction
from undefined.API import trace
import unittest



class TestTrace(unittest.TestCase):

    def setUp(self):
        self.f1 = lambda x: sqrt(exp(sin(x)))
        self.f2 = lambda x: exp(x**2 + sin(x))
        self.f3 = lambda x: sin(exp(2**x))
        self.f4 = lambda x: x**sin(x+2)
        self.f5 = lambda x, y: exp(1-6*x) * tan(4*x + 2*y)
        self.f6 = lambda x: 1 / (8 + cos(2*np.pi*x))
        self.f7 = lambda x, y, z: x*exp(y**2 - z**2)
        self.f8 = lambda x: log(x*3 + tan(2*x), np.e) + 2**x
        self.f9 = lambda x, y: x**y
        self.f10 = lambda x: log(x*3, np.e)
        self.f11 = lambda x: cos(exp(2*x))
        self.f12 = lambda x, y: log(1 + 6*x) * tan(4*x + 2*y)
        self.f13 = lambda x: log(6*x, 10) * tan(4*x)
        self.f1000 = lambda x, y: exp(1-6*x) * tan(4*x + 2*y) + x**2*y
        self.f1001 = lambda x, y: exp(1-6*x) / tan(4*x + 2*y) + x**2*y
        self.f1002 = lambda x, y: exp(1-6*x) // tan(4*x + 2*y) + x**2*y
        self.f1003 = lambda x: x//192
        self.f1004 = lambda x: x/192
        self.f1005 = lambda x: x**2 + x**5
        self.f1006 = lambda x: 2**x + 5**x
        self.f1007 = lambda x, y: x**y

    def assertNumpyArraysEqual(self, o1, o2):
        if o1.shape != o2.shape:
            raise AssertionError("Shapes don't match")
        if not np.allclose(o1, o2):
            raise AssertionError("Elements don't match!")

    def tearDown(self):
        pass

    def test_forward(self):
        self.assertEqual(trace(self.f1, x=2), (1.58, -0.328))
        self.assertEqual(trace(self.f2, x=3), (9331.210, 46749.434))
        self.assertEqual(trace(self.f3, x=3), (0.40, -15137.189))
        self.assertEqual(trace(self.f4, x=3), (0.35, -0.003))

        self.assertEqual(trace(self.f5, x=2, y=3), (0.0, [[0.003], [0.002]]))
        self.assertEqual(trace(self.f7, x=1, y=2, z=3),
                         (0.01, [[0.007],  [0.027], [-0.04]]))
        self.assertEqual(trace(self.f8, x=3), (10.16, 6.139))
        self.assertEqual(trace(self.f9, x=5, y=3), (125, [[75.], [201.18]]))
        self.assertEqual(trace(self.f10, x=2), (1.79, 0.5))

        def f22(x):
            return x+cos(2*x)
        with self.assertRaises(TypeError):
            trace(f22(2), x=2)


    def test_forward_trace(self):
        result1 = trace(self.f1, x=np.array([[2]]))
        self.assertEqual(result1, (np.array([[1.58]]), np.array([[-0.328]])))
        self.assertEqual(trace(self.f2, x=np.array(
            [[3]])), (np.array([[9331.21]]), np.array([[46749.434]])))
        self.assertEqual(trace(self.f8, x=np.array([[3]])), (np.array(
            [[10.16]]), np.array([[6.139]])))  # raise error
        self.assertEqual(trace(self.f11, x=np.array(
            [[2]])), (np.array([[-0.37]]), np.array([[101.418]])))
        self.assertEqual(trace(self.f13, x=np.array(
            [[3]])), (np.array([[-0.8]]), np.array([[6.959]])))
        self.assertEqual(trace(self.f13, x=np.array(
            [[3]])), (np.array([[-0.8]]), np.array([[6.959]])))
        with self.assertRaises(TypeError):
            trace(self.f1, x=np.array([]))

        # test multiple x values at the same time
        self.assertNumpyArraysEqual(trace(self.f1, x=np.array([[1, 2]]))[
                                    0], np.array([[1.52, 1.58]]))

        self.assertEqual(trace(self.f1, x=np.array([[1, 2]]))[
                         1], [[0.411, -0.328]])

        with self.assertRaises(TypeError):
            trace(self.f1, x="2")

        self.assertNumpyArraysEqual(trace([self.f1, self.f2], x=2)[
                                    0], np.array([1.58, 135.54]))
        self.assertNumpyArraysEqual(trace([self.f1, self.f2], x=2)[
                                    1], np.array([-3.2800e-01,  4.8577e+02]))

        g = UDGraph(np.array([[2], [1]]))
        with self.assertRaises(ValueError):
            g.__rpow__(UDGraph(np.array([[np.pi/2]])))

        with self.assertRaises(TypeError):
            g.__rpow__("str")

        g = UDGraph(np.array([[2], [1]]))
        g.__hash__()

    def test_reverse(self):
        g = UDGraph(9999.909)
        g2 = g.__radd__(UDGraph(np.pi/2))
        GeneratorHelper._radd(g2, g, seed_dic={})

        g = UDGraph(999.09)
        g2 = g.__rmul__(UDGraph(np.pi/2))
        GeneratorHelper._rmul(g2, g, seed_dic={})

        g = UDGraph(np.pi/2)
        g2 = g.__neg__()
        GeneratorHelper._neg(g2, g, seed_dic={})

        g = UDGraph(9999.909)
        g2 = g.__rsub__(UDGraph(np.pi/2))
        GeneratorHelper._rsub(g2, g, seed_dic={})

        g = UDGraph(9999.909)
        g2 = g.__rtruediv__(UDGraph(np.pi/2))
        GeneratorHelper._rtruediv(g2, g, seed_dic={})

        g = UDGraph(9999.909)
        g2 = g.__rfloordiv__(UDGraph(np.pi/2))
        GeneratorHelper._rfloordiv(g2, g, seed_dic={})

        g = UDGraph(3.5)
        g2 = g.__rpow__(UDGraph(np.pi/2))
        GeneratorHelper._rpow(g2, g, seed_dic={})

        g = UDGraph(3.5)
        g2 = g.__rpow__(UDGraph(np.array([[np.pi/2]])))
        GeneratorHelper._rpow(g2, g, seed_dic={})

        g = UDGraph(np.array(3.5))
        g2 = g.__rpow__(UDGraph(3.5))
        GeneratorHelper._rpow(g2, g, seed_dic={})

        g1 = UDGraph(np.array([[3.5]]))
        g2 = g1.__rpow__(UDGraph(np.array([[np.pi/2]])))
        GeneratorHelper._rpow(g2, g1, seed_dic={})


        self.assertEqual(repr(g), "Computational Graph (3.5, UDPrimitive.VAR)")

        result1001 = trace(self.f1001, mode="reverse", x=np.pi, y=998)

        self.assertEqual(result1001, (9849.87, [[6270.619], [9.87]]))

        result1002 = trace(self.f1002, mode="reverse", x=np.pi, y=998)

        self.assertEqual(result1002, (9849.87, [[6269.619], [8.87]]))

        result1003 = trace(self.f1003, mode="reverse", x=1343)

        self.assertEqual(result1003, (6, 0))

        result1004 = trace(self.f1004, mode="reverse", x=1343)

        self.assertEqual(result1004, (6.99, 0.005))

        result1005 = trace(self.f1005, mode="reverse", x=3)

        self.assertEqual(result1005, (252, 411))

        result1006 = trace(self.f1006, mode="reverse", x=5)

        self.assertEqual(result1006, (3157, 5051.674))

        result1007 = trace(self.f1005, mode="reverse", x=np.array([[1343]]))

        self.assertEqual(result1005, (252, 411))

        result1006 = trace(self.f1006, mode="reverse", x=np.array([[3]]))

        self.assertEqual(result1006, (np.array([[133]]), 206.725))

        result1007 = trace(self.f1007, mode="reverse",
                           x=np.array([[3]]), y=np.array([[3]]))

        self.assertEqual(result1007, (np.array(
            [[27]]), [[27.0], [29.663]]), 206.725)

        result1007 = trace(self.f1007, mode="reverse", x=3, y=np.array([[3]]))

        self.assertEqual(result1007, (np.array(
            [[27]]), [[27.0], [29.663]]), 206.725)

        result1007 = trace(self.f1007, mode="reverse", x=np.array([[3]]), y=3)

        self.assertEqual(result1007, (np.array(
            [[27]]), [[27.0], [29.663]]), 206.725)

        result1007 = trace(self.f1007, mode="reverse", x=3, y=3)

        self.assertEqual(result1007, (np.array(
            [[27]]), [[27.0], [29.663]]), 206.725)
        result1 = trace(self.f1, mode="reverse", x=2)

        self.assertEqual(result1, (1.58, [-0.328]))

        result2 = trace(self.f2, mode="reverse", x=2)
        self.assertEqual(result2, (135.54, [485.77]))

        self.function_ = lambda x: x+1

        result3 = trace(self.function_, mode="reverse", x=2)
        self.assertEqual(result3, (3, [1]))

        self.function_2 = lambda x: 3*x - 1
        result4 = trace(self.function_2, mode="reverse", x=2)
        self.assertEqual(result4, (5, [3]))

        self.function_3 = lambda x: 1 - x*3
        result5 = trace(self.function_3, mode="reverse", x=2)
        self.assertEqual(result5, (-5, [-3]))

        self.function_4 = lambda x: 1 + x

        result6 = trace(self.function_4, mode="reverse", x=2)
        self.assertEqual(result6, (3, [1]))

        result7 = trace(self.f5, mode="reverse", x=2, y=1)
        self.assertEqual(result6, (3, [1]))

        result8 = trace(self.f6, mode="reverse", x=2)
        self.assertEqual(result8, (0.11, [-0.0]))

        result9 = trace(self.f7, mode="reverse", x=2, y=1, z=3)
        self.assertEqual(result6, (3, [1]))

        # result10 = trace(self.f8,mode = "reverse",x = 2)
        # self.assertEqual(result10,(3, [1]))

        # result11 = trace(self.f9,mode = "reverse",x = 2,y =1)
        # self.assertEqual(result11,(3, [1]))

        result12 = trace(self.f10, mode="reverse", x=2)
        self.assertEqual(result12, (1.79, [0.5]))

        result13 = trace(self.f11, mode="reverse", x=2)
        self.assertEqual(result13, (-0.37, [101.418]))

        result14 = trace(self.f12, mode="reverse", x=2, y=1)
        self.assertEqual(result14, (1.66, [[14.872], [7.286]]))

        result15 = trace(self.f13, mode="reverse", x=2)
        self.assertEqual(result15, (-7.34, [[202.429]]))

        self.f100 = lambda x: 1+x**2
        result16 = trace(self.f100, mode="reverse", x=2)
        self.assertEqual(result16, (5, [4]))

        self. f101 = lambda x, y: x*y
        result17 = trace(self.f101, mode="reverse", x=2, y=1)
        self.assertEqual(result17, (2, [[1], [2]]))

        self.f102 = lambda x, y: -x*y
        result18 = trace(self.f102, mode="reverse", x=2, y=1)
        self.assertEqual(result18, (-2, [[-1], [-2]]))

        self.f103 = lambda x: - x + 1
        result19 = trace(self.f103, mode="reverse", x=2)
        self.assertEqual(result19, (-1, [-1]))

        self.fcs = lambda x: cos(x)+1
        result19 = trace(self.fcs, mode="reverse", x=np.array([[2]]))
        self.assertEqual(result19, (np.array([[0.58]]), -0.909))

        with self.assertRaises(TypeError):
            trace(self.fcs, mode="reverse", x="str")

        self.sine = lambda x: sin(x)+1
        result19 = trace(self.sine, mode="reverse", x=np.array([[20]]))
        self.assertEqual(result19, (np.array([[1.91]]), 0.408))

        with self.assertRaises(TypeError):
            trace(self.sine, mode="reverse", x="str")

        self.tange = lambda x: sin(x)+1
        result19 = trace(self.tange, mode="reverse", x=np.array([[50]]))
        self.assertEqual(result19, (np.array([[0.74]]), 0.965))

        with self.assertRaises(TypeError):
            trace(self.tange, mode="reverse", x="str")

    def test_multiple_function(self):
        def f1(x): return sqrt(exp(sin(x)))
        def f2(x): return 2*x + sqrt(x)

        self.assertEqual(trace([f1, f2], x=np.array([[1, 2]]))[
                         1].tolist(), [[[0.411, -0.328]], [[2.5,  2.354]]])
        self.assertEqual(trace([f1, f2], x=np.array([[1, 2, 3]]))[0].tolist(), [
                         [[1.52, 1.58, 1.07]], [[3., 5.41, 7.73]]])
        self.assertEqual(trace([f1, f2], x=np.array([[1, 2, 3]]))[1].tolist(), [
                         [[0.411, -0.328, -0.531]], [[2.5,  2.354,  2.289]]])

    def test_multiple_variable_function(self):

        def f3(x, y): return x**2 + 2**y
        def f4(x, y): return 2*x - 2/y

        self.assertEqual(trace([f3, f4], x=np.array([[1, 2]]), y=4)[
                         0].tolist(), [[[17., 20.]], [[1.5,  3.5]]])
        self.assertEqual(trace([f3, f4], x=np.array([[1, 2]]), y=4)[1].tolist(), [
                         [[2.,  4.], [11.09, 11.09]], [[2.,  2.], [0.125,  0.125]]])

        self.assertEqual(trace([f3, f4], mode="reverse", x=np.array([[1, 2]]), y=4)[
                         0].tolist(), [[[17., 20.]], [[1.5,  3.5]]])
        self.assertEqual(trace([f3, f4], mode="reverse", x=np.array([[1, 2]]), y=4)[
                         1].tolist(), [[[2.,  4.], [11.09, 11.09]], [[2.,  2.], [0.125,  0.125]]])

        self.assertEqual(trace([f3, f4], x=4, y=np.array([[1, 2]]))[
                         0].tolist(), [[[18., 20.]], [[6.,  7.]]])
        self.assertEqual(trace([f3, f4], x=4, y=np.array([[1, 2]]))[1].tolist(), [
                         [[8., 8.], [1.386, 2.773]], [[2., 2.], [2., 0.5]]])

        self.assertEqual(trace([f3, f4], mode="reverse", x=4, y=np.array([[1, 2]]))[
                         0].tolist(), [[[18., 20.]], [[6.,  7.]]])
        self.assertEqual(trace([f3, f4], mode="reverse", x=4, y=np.array([[1, 2]]))[
                         1].tolist(), [[[8., 8.], [1.386, 2.773]], [[2., 2.], [2., 0.5]]])


if __name__ == "__main__":
    unittest.main()



import sys
sys.path.append(
    "./src/")
import unittest
from undefined.API import trace
from undefined.Calculator import *



class TestAPI(unittest.TestCase):

    def setUp(self):
        self.a = 2
        self.b = - 1.11
        self.c = 99999.99
        self.d = -99999.99
        self.e = np.array([[self.a, self.b]])
        self.f = np.array([[self.a, self.b, self.c, self.d]])
        self.g = math.pi / 2
        self.h = math.e
        self.j = np.array([[self.a, self.b, self.g, self.h]])
        self.s = "str"

        self.f0 = lambda x: sqrt(x)
        self.f1 = lambda x: x + 2
        self.f2 = lambda x, y: x + y - 1
        self.f3 = lambda x, y: 2*x + sqrt(y)
        self.f12 = lambda x, y: sqrt(x) + sqrt(y)

    def test_trace_forward(self):
        # Stop using string comparing!!!!!!!
        self.assertEqual(trace(self.f1, x=2), (4, 1))
        self.assertEqual(trace(self.f1, mode='forward', x=2), (4, 1))

        self.assertEqual(trace(self.f2, x=2, y=4), (5, [[1], [1]]))
        self.assertEqual(trace(self.f2, y=4, x=2), (5, [[1], [1]]))
        self.assertEqual(trace(self.f3, x=2, y=4), (6.0, [[2.], [0.25]]))
        self.assertEqual(trace(self.f3, x=2, y=4, seeds=np.array(
            [[2, 2], [1, 1]])), (6.0, [[4.25], [4.25]]))

        with self.assertRaises(TypeError):
            trace(self.f1, seeds="1", x=np.array([[999]]))

        with self.assertRaises(TypeError):
            trace(self.f2, seeds="1", x=np.array([[999]]), y=np.array([[99]]))

        with self.assertRaises(TypeError):
            trace(self.f2, seeds=np.array([1, 2]), x=np.array(
                [[999]]), y=np.array([[99]]))

        self.assertEqual(trace(self.f12, seeds=np.array(
            [[1, 2], [0, 1]]), x=1, y=2)[1], [[0.5], [1.354]])
        self.assertEqual(trace(self.f0, seeds=1, x=np.array([[10, 1]]))[
                         1], [[0.158, 0.5]])

    def test_trace_with_incompatible_inputs(self):
        with self.assertRaises(AttributeError):
            trace(self.f2, mode='undefined', x=np.array(
                [[1, 2, 3]]), y=np.array([[1, 2]]))
        with self.assertRaises(AttributeError):
            trace(self.f2, x=np.array(
                [[1, 2, 3]]), y=np.array([[1, 2]]))
        with self.assertRaises(AttributeError):
            trace(self.f2, x=1, y=2, z=3)

    def test_trace_with_incompatible_seeds(self):
        with self.assertRaises(AttributeError):
            trace(self.f2, seeds=np.array([[1, 3, 2], [1, 3, 2], [2, 3, 3]]), x=np.array(
                [[2, 3]]), y=np.array([[1, 2]]))

    def test_trace_with_incompatible_seeds_reverse(self):
        with self.assertRaises(AttributeError):
            trace(self.f2, seeds=np.array([[1, 3, 2], [1, 3, 2], [2, 3, 3]]), mode="reverse", x=np.array(
                [[2, 3]]), y=np.array([[1, 2]]))

    def test_trace_with_scalar_inputs_seeds(self):
        with self.assertRaises(TypeError):
            trace(self.f1, seeds=np.array([[1, 0]]), x=1)
        self.assertEqual(trace(self.f1, seeds=2, x=1), (3, 2))
        self.assertEqual(trace(self.f2, seeds=np.array(
            [[1, 0], [0, 1]]), x=2, y=999)[1], [[1], [1]])
        with self.assertRaises(TypeError):
            trace(self.f2, seeds=1, x=2, y=999)
        with self.assertRaises(TypeError):
            trace(self.f2, seeds="seed", x=2, y=999)
        with self.assertRaises(TypeError):
            trace(self.f2, seeds=np.array([[1, 0], [0, 1]]), x="2", y=999)
        with self.assertRaises(AttributeError):
            trace(self.f2, seeds=np.array([[1], [0]]), x=2, y=999)

    def test_trace_with_scalar_inputs_seeds_reverse(self):
        with self.assertRaises(TypeError):
            trace(self.f1, seeds=np.array([[1, 0]]), mode='reverse', x=1)
        self.assertEqual(trace(self.f1, seeds=2, mode='reverse', x=1), (3, 2))
        self.assertEqual(trace(self.f2, seeds=np.array(
            [[1, 0], [0, 1]]), mode='reverse', x=2, y=999)[1], [[1], [1]])
        with self.assertRaises(TypeError):
            trace(self.f2, seeds=1, mode='reverse', x=2, y=999)
        with self.assertRaises(TypeError):
            trace(self.f2, seeds="seed", mode='reverse', x=2, y=999)
        with self.assertRaises(TypeError):
            trace(self.f2, seeds=np.array(
                [[1, 0], [0, 1]]), mode='reverse', x="2", y=999)
        with self.assertRaises(AttributeError):
            trace(self.f2, seeds=np.array(
                [[1], [0]]), mode='reverse', x=2, y=999)

    def test_trace_with_vector_inputs_seeds(self):
        self.assertEqual(trace(self.f2, seeds=np.array([[1, 0], [0, 1]]), x=np.array(
            [[2, 3]]), y=np.array([[1, 2]]))[1], [[1, 1], [1, 1]])

    def test_trace_with_vector_inputs_seeds_reverse(self):
        self.assertEqual(trace(self.f2, seeds=np.array([[1, 0], [0, 1]]), mode='reverse', x=np.array(
            [[2, 3]]), y=np.array([[1, 2]]))[1], [[1, 1], [1, 1]])

    def test_trace_with_different_moded(self):
        self.assertEqual(trace(self.f1, x=2), (4, 1))
        self.assertEqual(trace(self.f1, mode='forward', x=2), (4, 1))
        self.assertEqual(trace(self.f1, mode='reverse', x=2), (4, 1))
        with self.assertRaises(AttributeError):
            trace(self.f1, mode='undefined', x=2)

    def test_trace_reverse(self):
        self.assertEqual(trace(self.f1, mode='reverse', x=2), (4, 1))
        self.assertEqual(trace(self.f1, mode='reverse', x=2), (4, 1))
        with self.assertRaises(AttributeError):
            trace(self.f1, mode='undefined', x=2)
        self.assertEqual(trace(self.f2, mode='reverse',
                         x=2, y=4), (5, [[1], [1]]))
        self.assertEqual(trace(self.f2, mode='reverse',
                         y=4, x=2), (5, [[1], [1]]))
        self.assertEqual(trace(self.f3, mode='reverse',
                         x=2, y=4), (6.0, [[2.], [0.25]]))

        self.assertEqual(str(trace(self.f1, x=2)), '(4, 1)')
        self.assertEqual(str(trace(self.f1, mode='forward', x=2)), '(4, 1)')

        with self.assertRaises(AttributeError) as context:
            trace(self.f1, mode='undefined', x=2)

        with self.assertRaises(TypeError) as context:
            (trace(self.f1, mode='reverse', x=np.array([2, 1])))

        with self.assertRaises(TypeError) as context:
            (trace(self.f1, mode='reverse', x="1"))

    def test_trace_multiple_vector_inputs(self):
        self.assertEqual(trace(self.f3, x=np.array([[2, 2]]), y=np.array([[4, 4]]))[
                         1], [[2., 2.], [0.25, 0.25]])
        self.assertEqual(trace(self.f3, mode='reverse', x=np.array(
            [[2, 2]]), y=np.array([[4, 4]]))[1], [[2., 2.], [0.25, 0.25]])

    def test_trace_single_vector_input(self):
        self.assertEqual(trace(self.f1, x=np.array([[2, 2]]))[1], [[1, 1]])
        with self.assertRaises(TypeError):
            trace(self.f1, x=np.array([]))

    def test_trace_non_lambda_function(self):
        with self.assertRaises(TypeError):
            trace("Function", x=1)

    def test_trace_vector_functions(self):
        self.assertEqual(trace([self.f2, self.f3], x=2, y=4)[
                         1].tolist(), [[[1.0], [1.0]], [[2.0], [0.25]]])

    def test_mixed_inputs(self):
        self.assertEqual(trace(self.f3, x=np.array([[2, 2]]), y=4)[
                         0].tolist(), [[6, 6]])
        self.assertEqual(trace(self.f3, mode='reverse', x=np.array([[2, 2]]), y=4)[
                         0].tolist(), [[6, 6]])

        self.assertEqual(trace(self.f3, x=np.array([[2, 2]]), y=4)[
                         1], [[2., 2.], [0.25, 0.25]])
        self.assertEqual(trace(self.f3, mode='reverse', x=np.array([[2, 2]]), y=4)[
                         1], [[2., 2.], [0.25, 0.25]])
        self.assertEqual(trace(self.f3, mode='reverse', x=np.array([[2, 2]]), y=np.array([[4, 4]]))[
                         1], [[2., 2.], [0.25, 0.25]])


if __name__ == "__main__":
    unittest.main()



import sys
from typing import Type
import numpy as np
sys.path.append("./src/")
from undefined.Calculator import sin, cos, exp
from undefined.GraphGenerator import GraphGenerator, UDGraph
import unittest

class TestUDGraph(unittest.TestCase):

    # the setUp and tearDown are the methods to keep the code dry.
    def setUp(self):

        alpha = 2.0
        beta = 3.0

        a = 2.0
        x = UDGraph(a)

        # addition
        self.f1 = -2*x + beta
        self.f2 = beta + 2*x
        self.f3 = 5*x + 3*x
        self.f4 = x + 5*x + 3*x
        self.f5 = 5 + 2

        # substraction
        self.f6 = 2*x - beta
        self.f7 = beta - 2*x
        self.f8 = 5*x - 3*x
        self.f9 = x - 5*x - 3*x
        self.f10 = 5 - 2

        # multiplication
        self.f11 = 2*x + beta
        self.f12 = x*6 + beta
        self.f13 = beta + x*2
        self.f14 = beta + 4*x
        self.f15 = (x + 1) * (x - 2)

        # true divide
        self.f16 = 3*x/2
        self.f17 = (x+1)/(x-5)
        self.f20 = 4/3*x
        self.f24 = 343 / x

        # floor divide
        self.f18 = 3*x//4
        self.f19 = (x+1)//(x-5)
        self.f21 = 4//3*x
        self.f25 = 343 // x

        # power
        self.f22 = 4*x**3
        self.f23 = 4**2
        self.f50 = exp(sin(x)) - cos(x**0.5) * sin((cos(x)**2.0 + x**2.0)**0.5)
        self.f60 = 2**x
        self.gg = GraphGenerator(self.f3, {"x": 1})

    # tearDown is used to delete output files if applicable.

    def tearDown(self):
        pass

    # the convention is to name the testing function as "test_<meaningful name>",
    # otherwise the unittest will not run.
    def test_add(self):

        a = 2.0
        x = UDGraph(a)

        self.assertEqual(self.f1.val, -1.0)
        # self.assertEqual(self.f1.der, -2.0)

        self.assertEqual(self.f2.val, 7.0)
        # self.assertEqual(self.f2.der, 2.0)

        self.assertEqual(self.f3.val, 16.0)
        # self.assertEqual(self.f3.der, 8.0)

        self.assertEqual(self.f4.val, 18.0)
        # self.assertEqual(self.f4.der, 9.0)

        with self.assertRaises(Exception):
            self.f5.val
        # with self.assertRaises(Exception):
            # self.f5.der

        with self.assertRaises(Exception):
            self.f30 = "1" + 2*x

        with self.assertRaises(Exception):
            self.f40 = 2*x + "1"

    def test_radd(self):
        a = 2.0
        x = UDGraph(a)
        y = UDGraph(a + 2)
        z = "1"
        a = y.__radd__(x)
        self.assertEqual(a.val, 6)
        # self.assertEqual(a.der, 2)
        with self.assertRaises(Exception):
            x.__radd__(z)

    def test_sub(self):

        a = 2.0
        x = UDGraph(a)

        self.assertEqual(self.f6.val, 1.0)
        # self.assertEqual(self.f6.der, 2.0)

        self.assertEqual(self.f7.val, -1.0)
        # self.assertEqual(self.f7.der, -2.0)

        self.assertEqual(self.f8.val, 4.0)
        # self.assertEqual(self.f8.der, 2.0)

        self.assertEqual(self.f9.val, -14.0)
        # self.assertEqual(self.f9.der, -7.0)

        with self.assertRaises(Exception):
            self.f10.val
        # with self.assertRaises(Exception):
        # 	self.f10.der

        with self.assertRaises(Exception):
            self.f31 = "1" - 2*x

        with self.assertRaises(Exception):
            self.f41 = 2*x - "1"

    def test_rsub(self):
        a = 2.0
        x = UDGraph(a)
        y = UDGraph(a + 2)
        z = "1"
        a = y.__rsub__(x)
        self.assertEqual(a.val, -2)
        # self.assertEqual(a.der, 0)
        with self.assertRaises(Exception):
            x.__rsub__(z)

    def test_mul(self):

        a = 2.0
        x = UDGraph(a)

        self.assertEqual(self.f11.val, 7.0)
        # self.assertEqual(self.f11.der, 2.0)

        self.assertEqual(self.f12.val, 15.0)
        # self.assertEqual(self.f12.der, 6.0)

        self.assertEqual(self.f13.val, 7.0)
        # self.assertEqual(self.f13.der, 2.0)

        self.assertEqual(self.f14.val, 11.0)
        # self.assertEqual(self.f14.der, 4.0)

        with self.assertRaises(Exception):
            self.f32 = "2" * 4*x

        with self.assertRaises(Exception):
            self.f42 = 4*x * "2"

    def test_rmul(self):
        a = 2.0
        x = UDGraph(a)
        y = UDGraph(a + 2)
        z = "1"
        a = y.__rmul__(x)
        self.assertEqual(a.val, 8)
        # self.assertEqual(a.der, 6)
        with self.assertRaises(Exception):
            x.__rmul__(z)

    def test_neg(self):
        a = 2.0
        x = UDGraph(a)

        self.assertEqual(str(x.__neg__()), 'RMUL\nValue:-2.0\n')

    def test_truedev(self):

        a = 2.0
        x = UDGraph(a)

        self.assertEqual(self.f16.val, 3)
        # self.assertEqual(self.f16.der, 1.5)

        self.assertEqual(self.f17.val, -1.0)
        # self.assertEqual(round(self.f17.der, 2), -0.67)

        self.assertEqual(np.round(self.f20.val, 2), 2.67)
        # self.assertEqual(round(self.f20.der, 2), 1.33)

        self.assertEqual(np.round(self.f24.val, 2), 171.5)
        # self.assertEqual(round(self.f24.der, 2), -85.75)

        with self.assertRaises(Exception):
            self.f35 = "2" / 4*x

        with self.assertRaises(Exception):
            self.f45 = 4*x / "2"

    def test_rtruediv(self):
        a = 2.0
        x = UDGraph(a)
        y = UDGraph(a + 2)
        z = "1"
        a = y.__rtruediv__(x)
        self.assertEqual(a.val, 0.5)
        # self.assertEqual(a.der, 0.125)
        with self.assertRaises(Exception):
            x.__rtruediv__(z)

    def test_floordev(self):

        a = 2.0
        x = UDGraph(a)

        self.assertEqual(self.f18.val, 1.0)
        # self.assertEqual(self.f18.der, 0)

        self.assertEqual(self.f19.val, -1.0)
        # self.assertEqual(round(self.f19.der, 2), -1.0)

        self.assertEqual(np.round(self.f21.val, 2), 2)
        # self.assertEqual(round(self.f21.der, 2), 1)

        self.assertEqual(np.round(self.f25.val, 2), 171)
        # self.assertEqual(round(self.f25.der, 2), -86)

        with self.assertRaises(Exception):
            self.f36 = "2" // 4*x

        with self.assertRaises(Exception):
            self.f46 = 4*x // "2"

    def test_rfloordev(self):
        a = 2.0
        x = UDGraph(a)
        y = UDGraph(a + 2)
        z = "1"
        a = y.__rfloordiv__(x)
        self.assertEqual(a.val, 0)
        # self.assertEqual(a.der, 0)
        with self.assertRaises(Exception):
            x.__rfloordiv__(z)

    def test_power(self):

        self.assertEqual(self.f22.val, 32)
        # self.assertEqual(self.f22.der, 48)

        self.assertEqual(np.round(self.f50.val, 2), 2.34)
        # self.assertEqual(round(self.f50.der, 2), -0.64) # Value check

        with self.assertRaises(Exception):
            self.f23.val
        # with self.assertRaises(Exception):
        #	self.f23.der

        # test other as ndarray
        a = 2.0
        x = UDGraph(a)
        y = UDGraph(np.array([a]))
        self.f51 = x**y
        self.assertEqual(self.f51.val, 4)
        # self.assertEqual(self.f51.der,6.773)

        # test self/other both ndarray
        x = UDGraph(np.array([a]))
        y = UDGraph(np.array([a]))

        self.f52 = x ** y
        self.assertEqual(self.f52.val[0], 4.0)
        # self.assertEqual(self.f52.der[0],6.773)

        # raise value error
        x = UDGraph(np.array([1]))
        y = UDGraph(np.array([1, 2]))

        with self.assertRaises(ValueError) as context:
            x ** y

        # self.assertEqual(self.f52.val[0],4)
        # self.assertEqual(self.f52.der[0],6.773)

        # self is array, other is int/float

        x = UDGraph(np.array(2))
        y = UDGraph(2)

        self.f53 = x ** y
        self.assertEqual(self.f53.val, 4)
        # self.assertEqual(self.f53.der,6.773)

        self.assertEqual(self.f60.val, 4)
        # self.assertEqual(self.f60.der, 2.773)

        # test __rpow__
        # This pa
        x = UDGraph(a)

        self.f60 = 2 ** x
        self.assertEqual(self.f60.val, 4)
        # self.assertEqual(self.f60.der,2.773)

    def test_eq(self):
        self.assertTrue(self.f22 == 32)
        self.assertTrue(self.f19 == -1)
        self.assertTrue(np.round(self.f25.val, 2) == 171)
        self.assertTrue(np.round(self.f20.val, 2) == 2.67)

        self.assertTrue(self.f22 == self.f22)
        self.assertTrue(self.f11 == self.f13)

        with self.assertRaises(TypeError):
            self.f11 == "self.f13"

    def test_ne(self):
        self.assertTrue(self.f19 != -12)
        self.assertTrue(np.round(self.f25.val, 2) != 161)
        self.assertTrue(self.f22 != self.f19)

        with self.assertRaises(TypeError):
            self.f11 != "self.f13"

    def test_lt(self):
        self.assertTrue(self.f19 < 12)
        self.assertTrue(np.round(self.f25.val, 2) < 200)
        self.assertTrue(self.f19 < self.f25)

        with self.assertRaises(TypeError):
            self.f25 < "self.f19"

    def test_gt(self):
        self.assertTrue(self.f19 > -12)
        self.assertTrue(np.round(self.f25.val, 2) > -200)
        self.assertTrue(self.f25 > self.f19)

        with self.assertRaises(TypeError):
            self.f25 > "self.f19"

    def test_le(self):
        self.assertTrue(self.f19 <= 12)
        self.assertTrue(np.round(self.f25.val, 2) <= 200)
        self.assertTrue(self.f19 <= self.f25)

        self.assertTrue(self.f22 <= 32)
        self.assertTrue(self.f19 <= -1)

        with self.assertRaises(TypeError):
            self.f25 <= "self.f19"

    def test_ge(self):
        self.assertTrue(self.f19 >= -12)
        self.assertTrue(np.round(self.f25.val, 2) >= -200)
        self.assertTrue(self.f25 >= self.f19)

        self.assertTrue(self.f22 >= 32)
        self.assertTrue(self.f19 >= -1)

        with self.assertRaises(TypeError):
            self.f25 >= "self.f19"

    def test_repr(self):
        self.assertEqual(repr(self.f3), "Computational Graph (16.0, UDPrimitive.ADD)\n" +
                        "|\n" +
                        "|<-(parent)-Computational Graph (10.0, UDPrimitive.RMUL)\n" +
                        "|      |\n" +
                        "|      |<-(parent)-Computational Graph (2.0, UDPrimitive.VAR)\n" +
                        "|\n" +
                        "|<-(parent)-Computational Graph (6.0, UDPrimitive.RMUL)\n" +
                        "|      |\n" +
                        "|      |<-(parent)-Computational Graph (2.0, UDPrimitive.VAR)")
	
    def test_generate_graph(self):
        self.gg.generate_graph()
        self.assertNotEqual(self.gg._nxgraph, None)

    def test_generate_derivative(self):
        with self.assertRaises(TypeError):
            self.gg.generate_derivative("y")

    def test_generate_str(self):
        self.assertEqual(self.gg.generate_str(), "Computational Graph (16.0, UDPrimitive.ADD)\n" +
                        "|\n" +
                        "|<-(parent)-Computational Graph (10.0, UDPrimitive.RMUL)\n" +
                        "|      |\n" +
                        "|      |<-(parent)-Computational Graph (2.0, UDPrimitive.VAR)\n" +
                        "|\n" +
                        "|<-(parent)-Computational Graph (6.0, UDPrimitive.RMUL)\n" +
                        "|      |\n" +
                        "|      |<-(parent)-Computational Graph (2.0, UDPrimitive.VAR)")



# this will help to run the unittest directly.
if __name__ == "__main__":
    unittest.main()



import sys
# # temp solution for directory.
sys.path.append("./src/")
import matplotlib.pyplot as plt
import networkx as nx
import math
import numpy as np
from undefined.Utils import UDPrimitive, time, check_division_by_zero, check_pow


class UDGraph:
    def __init__(self, val, func=UDPrimitive.VAR, varname="intermediate"):
        # TODO
        """
        This class is where we overload all the operators, which will be used to calculate the derivatives.

        Args:
            val (numeric or numpy ndarray): value of function
            der (int, optional): derivative of function. Defaults to 1.
        """
        self._val = val
        self._func = func
        self._varname = varname
        if hasattr(self._val, 'shape'):
            self._valshape = self._val.shape
        else:
            self._valshape = 1
        self._parents = []
        self._params = {}

    @property
    def val(self):
        # TODO
        """
        This is a decorator return rouded input self.val

        Returns:
            array: 4 decimal rounded input of self.value
        """
        if isinstance(self._val, float):
            return round(self._val, 2)
        elif isinstance(self._val, np.ndarray):
            return np.round(self._val, 2)
        else:
            return self._val

    def __str__(self):
        """return the results in format. 
        Used in building the computational graph in the reverse mode.

        Returns:
            formatted string with the value.
        """
        return f"{self._func}\n".replace("UDPrimitive.", "") + f"Value:{self.val}\n"

    def __repr__(self):
        """return the computational graph root.

        Returns:
            formatted string with the value for the computational graph root. 
        """
        res = f'Computational Graph ({self.val}, {self._func})'
        for parent in self._parents:
            res += '\n|\n|<-(parent)-' + \
                '\n|      '.join(repr(parent).split('\n'))
        return res

    def __add__(self, other):
        """
        This allows to do addition with UDFunction instances or scalar numbers, and calculate the value after taking the derivative. 
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            other (UDFunction or numeric): object to add with
        Returns:
            UDFunction: a new object with new_val and new_der
        """

        if isinstance(other, UDGraph):
            new_val = self._val + other._val
            new_func = UDPrimitive.ADD
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._parents.append(other)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val + other
            new_func = UDPrimitive.ADD
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __mul__(self, other):
        """
        This allows to do multification with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative. 
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            other (UDFunction or numeric): object to multiply with
        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            new_val = self._val * other._val
            new_func = UDPrimitive.MUL
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._parents.append(other)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val * other
            new_func = UDPrimitive.MUL
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __radd__(self, other):
        """
        This is called when int/float or UDFunction instances + an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to add with
        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            new_val = self._val + other._val
            new_func = UDPrimitive.RADD
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(other)
            udgraph._parents.append(self)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val + other
            new_func = UDPrimitive.RADD
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __rmul__(self, other):
        """
        This is called when int/float or UDFunction instances * an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to multiply with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            new_val = self._val * other._val
            new_func = UDPrimitive.RMUL
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(other)
            udgraph._parents.append(self)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val * other
            new_func = UDPrimitive.RMUL
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __neg__(self):
        """
        This allows to negate UDFunction instances itself.

        Returns:
            UDFunction: object with neg value
        """
        return -1 * self

    def __sub__(self, other):
        """
        This allows to do subtraction with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative. 
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            other (UDFunction or numeric): object to subtract with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            new_val = self._val - other._val
            new_func = UDPrimitive.SUB
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._parents.append(other)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = self._val - other
            new_func = UDPrimitive.SUB
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __rsub__(self, other):
        """
        This is called when int/float or UDFunction instances - an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to subtract with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            new_val = other._val - self._val
            new_func = UDPrimitive.RSUB
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(other)
            udgraph._parents.append(self)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = other - self._val
            new_func = UDPrimitive.RSUB
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __truediv__(self, other):
        """
        This allows to do true division with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative. 
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            other (UDFunction or numeric): object to (true) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            check_division_by_zero(other._val)
            new_val = self._val / other._val
            new_func = UDPrimitive.TRUEDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._parents.append(other)
        elif isinstance(other, (int, float, np.ndarray)):
            check_division_by_zero(other)
            new_val = self._val / other
            new_func = UDPrimitive.TRUEDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __rtruediv__(self, other):
        """
        This is called when int/float or UDFunction instances / (divide) an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to (true) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        check_division_by_zero(self._val)
        if isinstance(other, UDGraph):
            new_val = other._val / self._val
            new_func = UDPrimitive.RTRUEDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(other)
            udgraph._parents.append(self)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = other / self._val
            new_func = UDPrimitive.RTRUEDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __floordiv__(self, other):  # self // other
        """
        This allows to do floor division with UDFunction instances or scalar numbers, , and calculate the value after taking the derivative.
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            other (UDFunction or numeric): object to (floor) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        if isinstance(other, UDGraph):
            check_division_by_zero(other._val)
            new_val = self._val // other._val
            new_func = UDPrimitive.FLOORDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._parents.append(other)
        elif isinstance(other, (int, float, np.ndarray)):
            check_division_by_zero(other)
            new_val = self._val // other
            new_func = UDPrimitive.FLOORDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __rfloordiv__(self, other):
        """
        This is called when int/float or UDFunction instances // (floor divide) an instance of Variable class.

        Args:
            other (UDFunction or numeric): object to (floor) divide with

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        check_division_by_zero(self._val)
        if isinstance(other, UDGraph):
            new_val = other._val // self._val
            new_func = UDPrimitive.RFLOORDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(other)
            udgraph._parents.append(self)
        elif isinstance(other, (int, float, np.ndarray)):
            new_val = other // self._val
            new_func = UDPrimitive.RFLOORDIV
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["constant"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __pow__(self, other):
        """
        This allows to do "to the power" with UDFunction instances or scalar numbers, and calculate the value after taking the derivative.
        ** operator.

        Args:
            degree (numeric): object to take power of.

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        new_func = UDPrimitive.POW
        if isinstance(other, UDGraph):
            check_pow(self._val, other._val)
            if isinstance(self._val, (int, float)):
                new_val = self._val ** other._val
            else:
                if isinstance(other._val, np.ndarray):
                    if other._val.shape[0] != self._val.shape[0]:
                        raise ValueError(
                            f"error raised by undefined: operands could not be broadcast together with shapes {other._val.shape} {self._val.shape}")
                    else:
                        new_val = self._val ** other._val
                else:
                    new_val = self._val ** other._val
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._parents.append(other)
        elif isinstance(other, (int, float, np.ndarray)):
            check_pow(self._val, other)
            if isinstance(self._val, np.ndarray):
                new_val = np.power(self._val, other)
            elif isinstance(self._val, (int, float)):
                new_val = self._val ** other
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["degree"] = other
        return udgraph

    def __rpow__(self, other):
        """
        This allows to do "to the power" with UDFunction instances or scalar numbers, and calculate the value after taking the derivative.
        ** operator.
        TypeError will raise if none of the self or other are UDFunction instances. 

        Args:
            degree (numeric): object to take power of.

        Returns:
            UDFunction: a new object with new_val and new_der
        """
        new_func = UDPrimitive.RPOW
        if isinstance(other, UDGraph):
            check_pow(other._val, self._val)
            if isinstance(other._val, (int, float)):
                new_val = other._val ** self._val
            else:
                if isinstance(self._val, np.ndarray):
                    if other._val.shape[0] != self._val.shape[0]:
                        raise ValueError(
                            f"error raised by undefined: operands could not be broadcast together with shapes {other._val.shape} {self._val.shape}")
                    else:
                        new_val = other._val ** self._val
                else:
                    new_val = other._val ** self._val
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(other)
            udgraph._parents.append(self)
        elif isinstance(other, (int, float, np.ndarray)):
            check_pow(other, self._val)
            new_val = other ** self._val
            udgraph = UDGraph(new_val, new_func)
            udgraph._parents.append(self)
            udgraph._params["base"] = other
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")
        return udgraph

    def __eq__(self, other):
        """compare whether the two UDGraph objects have the same values.
        Return true if equal, and false otherwise.

        raise TypeError is other is not a UDGraph object.

        Args:
            other ([UDGraph])

        Returns:
            True if equal. Otherwise False
        """
        if isinstance(other, UDGraph):
            return self.val == other.val
        elif isinstance(other, (int, float)):
            return self.val == other
        else:
            raise TypeError(
                "error raised by undefined: Need a UDGraph object to compare")

    def __ne__(self, other):
        """compare whether the two UDGraph objects have different values.
        raise TypeError is other is not a UDGraph object.

        Args:
            other ([UDGraph])

        Returns:
            True if not equal. Otherwise False
        """
        if isinstance(other, UDGraph):
            return self.val != other.val
        elif isinstance(other, (int, float)):
            return self.val != other
        else:
            raise TypeError(
                "error raised by undefined: Need a UDGraph object to compare")

    def __lt__(self, other):
        """overload the < operator
        raise TypeError is other is not a UDGraph object.

        Args:
            other ([UDGraph])

        Returns:
            True if less than. Otherwise False            
        """
        if isinstance(other, UDGraph):
            return self.val < other.val
        elif isinstance(other, (int, float)):
            return self.val < other
        else:
            raise TypeError(
                "error raised by undefined: Need a UDGraph object to compare")

    def __gt__(self, other):
        """overload the > operator
        raise TypeError is other is not a UDGraph object.

        Args:
            other ([UDGraph])

        Returns:
            True if greater than. Otherwise False
        """
        if isinstance(other, UDGraph):
            return self.val > other.val
        elif isinstance(other, (int, float)):
            return self.val > other
        else:
            raise TypeError(
                "error raised by undefined: Need a UDGraph object to compare")

    def __le__(self, other):
        """overload the > operator
        raise TypeError is other is not a UDGraph object.

        Args:
            other ([UDGraph])

        Returns:
            True if less than or equal to. Otherwise False
        """
        if isinstance(other, UDGraph):
            return self.val <= other.val
        elif isinstance(other, (int, float)):
            return self.val <= other
        else:
            raise TypeError(
                "error raised by undefined: Need a UDGraph object to compare")

    def __ge__(self, other):
        """overload the > operator
        raise TypeError is other is not a UDGraph object.

        Args:
            other ([UDGraph])

        Returns:
            True if greater than or equal to. Otherwise False
        """
        if isinstance(other, UDGraph):
            return self.val >= other.val
        elif isinstance(other, (int, float)):
            return self.val >= other
        else:
            raise TypeError(
                "error raised by undefined: Need a UDGraph object to compare")

    def __hash__(self):
        """hash function for udgraph

        Returns:
            int: hash value for udgraph
        """
        return hash((str(self._val), self._func))


class GeneratorHelper:
    @classmethod
    def _var(self, udgraph: UDGraph, variable: UDGraph, seed_dic):
        """This private helper function for UDGraph is used to get the values. This is used in the primitive.
        Generally speaking, this is the function to get node value in the computational graph. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            node value
        """
        if seed_dic and len(seed_dic) > 0:
            seed = seed_dic[udgraph._varname][variable._varname]
            if isinstance(udgraph._val, np.ndarray):
                return seed * np.ones(udgraph._valshape)
            else:
                return seed
        else:
            if variable._varname is not udgraph._varname:
                if isinstance(udgraph._val, np.ndarray):
                    return np.zeros(udgraph._valshape)
                else:
                    return 0
            else:
                if isinstance(udgraph._val, np.ndarray):
                    return np.ones(udgraph._valshape)
                else:
                    return 1

    @classmethod
    def _add(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for addition operation's derivative result. This is used in the primitive and __add__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' addition result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the addition derivative result. 
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]

            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) + GraphGenerator.function_dic[g2._func](g2, variable, seed_dic)

    @classmethod
    def _radd(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for addition operation's derivative result. This is used in the primitive and __radd__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' addition result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the right addition derivative result. 
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) + GraphGenerator.function_dic[g2._func](g2, variable, seed_dic)

    @classmethod
    def _mul(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for multiplication operation's derivative result. This is used in the primitive and __mul__() in UDGraph.
        If there is one parent node, return node multiplication result to the constant.
        Otherwise, return the two parental node multiplication result.

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the multiplication derivative result. 
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return udgraph._params["constant"] * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) * g2._val + GraphGenerator.function_dic[g2._func](g2, variable, seed_dic) * g1._val

    @classmethod
    def _rmul(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for multiplication operation's derivative result. This is used in the primitive and __rmul__() in UDGraph.
        If there is one parent node, return node multiplication result to the constant.
        Otherwise, return the two parental node multiplication result.

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the right multiplication derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return udgraph._params["constant"] * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) * g2._val + GraphGenerator.function_dic[g2._func](g2, variable, seed_dic) * g1._val

    @classmethod
    def _neg(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """ This is a helper function to convert the value to negative. Used in __ne__() function in UDGraph

        Args:
            udgraph (UDGraph): [description]
            variable (UDGraph): [description]

        Returns:
            the negative value of the node value.
        """
        g1 = udgraph._parents[0]
        return -1 * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)

    @classmethod
    def _sub(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for subtraction operation's derivative result. This is used in the primitive and __sub__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' subtraction result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the subtraction derivative result. 
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) - GraphGenerator.function_dic[g2._func](g2, variable, seed_dic)

    @classmethod
    def _rsub(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for subtraction operation's derivative result. This is used in the primitive and __rsub__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' subtraction result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the right subtraction derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return -1 * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) - GraphGenerator.function_dic[g2._func](g2, variable, seed_dic)

    @classmethod
    def _truediv(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for true division operation's derivative result. This is used in the primitive and __truediv__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' true division result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the true division derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) / udgraph._params["constant"]
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return (GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) * g2._val - GraphGenerator.function_dic[g2._func](g2, variable, seed_dic) * g1._val) / (g2._val * g2._val)

    @classmethod
    def _rtruediv(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for true division operation's derivative result. This is used in the primitive and __rtruediv__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' true division result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the right true division derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return -1 * udgraph._params["constant"] * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) / (g1._val * g1._val)
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return (GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) * g2._val - GraphGenerator.function_dic[g2._func](g2, variable, seed_dic) * g1._val) / (g2._val * g2._val)

    @classmethod
    def _floordiv(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for floor division operation's derivative result. This is used in the primitive and __floordiv__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' floor division result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the floor division derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) // udgraph._params["constant"]
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return (GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) * g2._val - GraphGenerator.function_dic[g2._func](g2, variable, seed_dic) * g1._val) // (g2._val * g2._val)

    @classmethod
    def _rfloordiv(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for floor division operation's derivative result. This is used in the primitive and __rfloordiv__() in UDGraph.
        If there is one parent node, return the parental node value
        Otherwise, return the two parental nodes' floor division result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the right floor division derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) // udgraph._params["constant"]
        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]
            return (GraphGenerator.function_dic[g1._func](g1, variable, seed_dic) * g2._val - GraphGenerator.function_dic[g2._func](g2, variable, seed_dic) * g1._val) // (g2._val * g2._val)

    @classmethod
    def _pow(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for power operation's derivative result. This is used in the primitive and __pow__() in UDGraph.
        If there is one parent node, return the parental node power derivative result
        Otherwise, return the two parental nodes' power derivative result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the power derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            if isinstance(g1._val, np.ndarray):
                return udgraph._params["degree"] * np.power(g1._val, udgraph._params["degree"] - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
            else:
                return udgraph._params["degree"] * g1._val ** (udgraph._params["degree"] - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)

        else:  # len == 2
            g1, g2 = udgraph._parents[0], udgraph._parents[1]

            if isinstance(g2._val, np.ndarray):
                if isinstance(g1._val, np.ndarray):
                    der1 = g2._val * np.power(
                        g1._val, g2._val - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
                    der2 = np.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2
                else:
                    der1 = g2._val * g1._val ** (
                        g2._val - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
                    der2 = math.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2
            else:  # g2._val is int
                if isinstance(g1._val, np.ndarray):
                    der1 = g2._val * \
                        np.power(g1._val, g2._val - 1) * \
                        GraphGenerator.function_dic[g1._func](
                            g1, variable, seed_dic)
                    der2 = np.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2
                else:
                    der1 = g2._val * g1._val ** (
                        g2._val - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
                    der2 = math.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2

    @classmethod
    def _rpow(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """This private helper function is used for power operation's derivative. This is used in the primitive and __rpow__() in UDGraph.
        If there is one parent node, return the parental node power derivative result
        Otherwise, return the two parental nodes' power derivative result

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            the right power derivative result.
        """
        if len(udgraph._parents) == 1:
            g1 = udgraph._parents[0]
            return math.log(udgraph._params["base"]) * udgraph._val * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            g1, g2 = udgraph._parents[0], udgraph._parents[1]

            if isinstance(g2._val, np.ndarray):
                if isinstance(g1._val, np.ndarray):
                    der1 = g2._val * np.power(
                        g1._val, g2._val - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
                    der2 = np.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2
                else:
                    der1 = g2._val * g1._val ** (
                        g2._val - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
                    der2 = math.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2
            else:  # g2._val is int
                if isinstance(g1._val, np.ndarray):
                    der1 = g2._val * \
                        np.power(g1._val, g2._val - 1) * \
                        GraphGenerator.function_dic[g1._func](
                            g1, variable, seed_dic)
                    der2 = np.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2
                else:
                    der1 = g2._val * g1._val ** (
                        g2._val - 1) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
                    der2 = math.log(g1._val) * udgraph._val * \
                        GraphGenerator.function_dic[g2._func](
                            g2, variable, seed_dic)
                    return der1 + der2

    @classmethod
    def _cos(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for cosine in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the cosine function

        Returns:
            cosine's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return - 1 * math.sin(g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return -1 * np.sin(g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _sin(self, udgraph: UDGraph, variable: UDGraph, seed_dic):
        """Calculate the derivative for sine in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the sine function

        Returns:
            sine's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return math.cos(g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return np.cos(g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _tan(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for tangent in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the tangent function

        Returns:
            tangent's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return (1 / (math.cos(g1._val)) ** 2) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return (1 / (np.cos(g1._val)) ** 2) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _arccos(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for arccosine in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the arccosine function

        Returns:
            arccosine's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return (-1 / math.sqrt(1 - g1._val**2)) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return (-1 / np.sqrt(1 - g1._val**2)) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _arcsin(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for arcsine in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the arcsine function

        Returns:
            arcsine's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return (1 / math.sqrt(1 - g1._val**2)) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return (1 / np.sqrt(1 - g1._val**2)) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _arctan(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for arctangent in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the arctangent function

        Returns:
            arctangent's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float, np.ndarray)):
            return (1 / (1 + g1._val**2)) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _sqrt(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for squared root in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the squared root function

        Returns:
            squared root's derivatives results calculated in reverse mode. 
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return 0.5 * math.pow(g1._val, -0.5) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return 0.5 * np.power(g1._val, -0.5) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _exp(self, udgraph: UDGraph, variable: UDGraph, seed_dic=None):
        """Calculate the derivative for exponential in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Raises:
            TypeError: if unlawful input to the exponential function

        Returns:
            exponential's derivatives results calculated in reverse mode.
        """
        g1 = udgraph._parents[0]
        if isinstance(g1._val, (int, float)):
            return math.exp(g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        elif isinstance(g1._val, np.ndarray):
            return np.exp(g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)
        else:
            raise TypeError(
                "error raised by undefined: unsupported attribute type.")

    @classmethod
    def _log(self, udgraph: UDGraph, variable: UDGraph, seed_dic):
        """Calculate the derivative for log in reverse mode. 

        Args:
            udgraph (UDGraph)
            variable (UDGraph)

        Returns:
            log's derivatives results calculated in reverse mode.
        """
        g1 = udgraph._parents[0]
        return 1 / (math.log(udgraph._params["base"]) * g1._val) * GraphGenerator.function_dic[g1._func](g1, variable, seed_dic)


class GraphGenerator:
    function_dic = {
        UDPrimitive.VAR: GeneratorHelper._var,
        UDPrimitive.ADD: GeneratorHelper._add,
        UDPrimitive.RADD: GeneratorHelper._radd,
        UDPrimitive.MUL:  GeneratorHelper._mul,
        UDPrimitive.RMUL:  GeneratorHelper._rmul,
        UDPrimitive.NEG:  GeneratorHelper._neg,
        UDPrimitive.SUB: GeneratorHelper._sub,
        UDPrimitive.RSUB: GeneratorHelper._rsub,
        UDPrimitive.TRUEDIV: GeneratorHelper._truediv,
        UDPrimitive.RTRUEDIV: GeneratorHelper._rtruediv,
        UDPrimitive.FLOORDIV: GeneratorHelper._floordiv,
        UDPrimitive.RFLOORDIV: GeneratorHelper._rfloordiv,
        UDPrimitive.POW: GeneratorHelper._pow,
        UDPrimitive.RPOW: GeneratorHelper._rpow,
        UDPrimitive.COS: GeneratorHelper._cos,
        UDPrimitive.SIN: GeneratorHelper._sin,
        UDPrimitive.TAN: GeneratorHelper._tan,
        UDPrimitive.SQRT: GeneratorHelper._sqrt,
        UDPrimitive.EXP: GeneratorHelper._exp,
        UDPrimitive.LOG: GeneratorHelper._log,
        UDPrimitive.ACOS: GeneratorHelper._arccos,
        UDPrimitive.ASIN: GeneratorHelper._arcsin,
        UDPrimitive.ATAN: GeneratorHelper._arctan,
    }

    def __init__(self, g, variables, seeds_dic=None):
        """initialize the input instances.
        """
        self._udgraph = g
        self._variables = variables
        self._nxgraph = None
        self._seeds_dic = seeds_dic

    def _generate_inner(self, g: UDGraph, nxgraph: nx.DiGraph):
        """build the computational graph
        """
        for parent in g._parents:
            nxgraph.add_edge(parent, g)
        for k, v in g._params.items():
            nxgraph.add_edge(f"{k}: {v}", g)
        for parent in g._parents:
            self._generate_inner(parent, nxgraph)

    def generate_graph(self):
        """
        generate the visualization graph

        """

        if not self._nxgraph:
            self._nxgraph = nx.DiGraph()
            self._generate_inner(self._udgraph, self._nxgraph)
        nx.draw_networkx(self._nxgraph, with_labels=True, font_size=6)
        plt.savefig(f"{time()}.png")

    def generate_str(self):
        """Output the computational graph in str and maintain the structure.
        """
        return repr(self._udgraph)

    def generate_derivative(self, var_name):
        """A wrapper function to generate the derivative given the input key value. 
        Used in the trace function. 
        """
        if var_name not in self._variables.keys():
            raise TypeError(
                "error raised by undefined: variable not defined in function")
        variable = self._variables[var_name]
        return np.round(GraphGenerator.function_dic[self._udgraph._func](self._udgraph, variable, self._seeds_dic), 3)
